# Parameters whose values may vary.

VARIABLE_PARAMS = set([
    'effective_search_space',
    'effective_hsp_length',
    'query',
    'query_length',
    'query_letters'
])


def checkCompatibleParams(initialParams, laterParams):
    """
    Check a later set of BLAST parameters against those originally found.

    @param initialParams: A C{dict} with the originally encountered BLAST
        parameter settings.
    @param laterParams: A C{dict} with BLAST parameter settings encountered
        later.
    @return: A C{str} summary of the parameter differences if the parameter
        sets differ, else C{None}.
    """

    # Note that although the params contains a 'date', its value is empty
    # (as far as I've seen). This could become an issue one day if it
    # becomes non-empty and differs between JSON files that we cat
    # together. In that case we may need to be more specific in our params
    # compatible checking.
    err = []
    for param in initialParams:
        if param in laterParams:
            if (param not in VARIABLE_PARAMS
                    and initialParams[param] != laterParams[param]):
                err.append(
                    '\tParam %r initial value %r differs from '
                    'later value %r' % (param, initialParams[param],
                                        laterParams[param]))
        else:
            err.append('\t%r found in initial parameters, not found '
                       'in later parameters' % param)
    for param in laterParams:
        if param not in initialParams:
            err.append('\t%r found in later parameters, not seen in '
                       'initial parameters' % param)

    return 'Summary of differences:\n%s' % '\n'.join(err) if err else None



from dark.blast.hsp import printHSP


def printBlastRecord(record):
    """
    Print a BLAST record.

    @param record: A BioPython C{Bio.Blast.Record.Blast} instance.
    """
    for key in sorted(record.__dict__.keys()):
        if key not in ['alignments', 'descriptions', 'reference']:
            print('%s: %r' % (key, record.__dict__[key]))
    print('alignments: (%d in total):' % len(record.alignments))
    for i, alignment in enumerate(record.alignments):
        print('  description %d:' % (i + 1))
        for attr in ['accession', 'bits', 'e', 'num_alignments', 'score']:
            print('    %s: %s' % (attr, getattr(record.descriptions[i], attr)))
        print('  alignment %d:' % (i + 1))
        for attr in 'accession', 'hit_def', 'hit_id', 'length', 'title':
            print('    %s: %s' % (attr, getattr(alignment, attr)))
        print('    HSPs (%d in total):' % len(alignment.hsps))
        for hspIndex, hsp in enumerate(alignment.hsps, start=1):
            print('      hsp %d:' % hspIndex)
            printHSP(hsp, '        ')



#!/usr/bin/env python

from __future__ import print_function

import sys
import tempfile

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description=('Given FASTA on stdin write the sequence ids and lengths '
                     'to stdout.'))

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    count = 0
    length = None
    with tempfile.TemporaryFile() as fp:
        for read in reads:
            count += 1
            if length is None:
                length = len(read)
            else:
                if len(read) != length:
                    raise ValueError(
                        'FASTA sequence %r was not of the '
                        'expected length (%d)' % (read.id, length))

            name = read.id.split()[0]
            fp.write(('%s %s\n' % (name, read.sequence)).encode('utf-8'))

        if count:
            # Print the phylip header with the number of sequences and their
            # common length.
            print('%d %d' % (count, length))
            fp.seek(0)
            write = sys.stdout.write
            while True:
                chunk = fp.read(1024)
                if chunk:
                    write(chunk.decode('utf-8'))
                else:
                    break



#!/usr/bin/env python

from __future__ import print_function

import sys

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions
from dark.utils import parseRangeString


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description=('Given FASTA on stdin and a set of filtering criteria '
                     'write filtered FASTA to stdout.'))

    parser.add_argument(
        '--quiet', action='store_true', default=False,
        help=('If True, do not print the final sequence summary.'))

    parser.add_argument(
        '--saveAs', default=None, choices=('fasta', 'fastq', 'fasta-ss'),
        help=('The output format. The default is to match the input format, '
              'so there is usually no need to specify this option. It can be '
              'used to force conversion from FASTQ to FASTA'))

    parser.add_argument(
        '--minLength', type=int,
        help='The minimum sequence length')

    parser.add_argument(
        '--maxLength', type=int,
        help='The maximum sequence length')

    parser.add_argument(
        '--removeGaps', action='store_true', default=False,
        help=("If True, gap ('-') characters in sequences will be removed."))

    parser.add_argument(
        '--whitelist', action='append', default=None,
        help='sequence titles (ids) that should be whitelisted')

    parser.add_argument(
        '--blacklist', action='append', default=None,
        help='sequence titles (ids) that should be blacklisted')

    parser.add_argument(
        '--titleRegex', default=None,
        help='a regex that sequence titles (ids) must match.')

    parser.add_argument(
        '--negativeTitleRegex', default=None,
        help='a regex that sequence titles (ids) must not match.')

    parser.add_argument(
        '--truncateTitlesAfter', default=None,
        help=('a string that sequence titles (ids) will be truncated beyond. '
              'If the truncated version of a title has already been seen, '
              'that title will be skipped.'))

    parser.add_argument(
        '--indices', type=int, action='append', default=None,
        help='sequence indices that should be returned (zero based)')

    parser.add_argument(
        '--head', type=int, default=None, metavar='N',
        help='only the first N sequences will be printed.')

    parser.add_argument(
        '--removeDuplicates', action='store_true', default=False,
        help=('If True, duplicate sequences will be removed. The first '
              'occurrence is kept.'))

    # See the docstring for dark.reads.Reads.filter for more detail on
    # randomSubset.
    parser.add_argument(
        '--randomSubset', type=int, default=None,
        help=('An integer giving the number of sequences that should be kept. '
              'These will be selected at random.'))

    # See the docstring for dark.reads.Reads.filter for more detail on
    # trueLength.
    parser.add_argument(
        '--trueLength', type=int, default=None,
        help=('The number of reads in the FASTA input. Only to be used with '
              'randomSubset'))

    parser.add_argument(
        '--sampleFraction', type=float, default=None,
        help=('A [0.0, 1.0] C{float} indicating a fraction of the reads that '
              'should be allowed to pass through the filter. The sample size '
              'will only be approximately the product of the sample fraction '
              'and the number of reads. The sample is taken at random.'))

    parser.add_argument(
        '--sequenceNumbersFile', default=None,
        help=('A file of (1-based) sequence numbers to retain. Numbers must '
              'be one per line.'))

    # A mutually exclusive group for either --keepIndices or --removeIndices.
    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        '--keepIndices',
        help=('Specify 1-based indices to keep. All other sequence indices '
              'will be removed. The indices must be given in the form e.g., '
              '24,100-200,260'))

    group.add_argument(
        '--keepIndicesFile',
        help=('Specify a file containing 1-based indices to keep. All other '
              'sequence indices will be removed. Lines in the file must be '
              'given in the form e.g., 24,100-200,260'))

    group.add_argument(
        '--removeIndices',
        help=('Specify 1-based indices to remove. All other sequence indices '
              'will be kept. The indices must be given in the form e.g., '
              '24,100-200,260'))

    group.add_argument(
        '--removeIndicesFile',
        help=('Specify a file containing 1-based indices to remove. All other '
              'sequence indices will be kept. Lines in the file must be given '
              'in the form e.g., 24,100-200,260'))

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    keepIndices = (
        parseRangeString(args.keepIndices, convertToZeroBased=True)
        if args.keepIndices else None)

    if args.keepIndicesFile:
        keepIndices = keepIndices or set()
        with open(args.keepIndicesFile) as fp:
            for lineNumber, line in enumerate(fp):
                try:
                    keepIndices.update(
                        parseRangeString(line, convertToZeroBased=True))
                except ValueError as e:
                    raise ValueError(
                        'Keep indices file %r line %d could not be parsed: %s'
                        % (args.keepIndicesFile, lineNumber, e))

    removeIndices = (
        parseRangeString(args.removeIndices, convertToZeroBased=True)
        if args.removeIndices else None)

    if args.removeIndicesFile:
        removeIndices = removeIndices or set()
        with open(args.removeIndicesFile) as fp:
            for lineNumber, line in enumerate(fp):
                try:
                    removeIndices.update(
                        parseRangeString(line, convertToZeroBased=True))
                except ValueError as e:
                    raise ValueError(
                        'Remove indices file %r line %d parse error: %s'
                        % (args.removeIndicesFile, lineNumber, e))

    reads.filter(
        minLength=args.minLength, maxLength=args.maxLength,
        removeGaps=args.removeGaps,
        whitelist=set(args.whitelist) if args.whitelist else None,
        blacklist=set(args.blacklist) if args.blacklist else None,
        titleRegex=args.titleRegex,
        negativeTitleRegex=args.negativeTitleRegex,
        truncateTitlesAfter=args.truncateTitlesAfter,
        indices=set(args.indices) if args.indices else None,
        head=args.head, removeDuplicates=args.removeDuplicates,
        randomSubset=args.randomSubset, trueLength=args.trueLength,
        sampleFraction=args.sampleFraction,
        sequenceNumbersFile=args.sequenceNumbersFile,
        keepIndices=keepIndices, removeIndices=removeIndices)

    saveAs = (
        args.saveAs or
        (args.fasta and 'fasta') or
        (args.fastq and 'fastq') or
        (args.fasta_ss and 'fasta-ss'))

    # Check for incompatible read/write formats. We can't write FASTQ
    # unless we have FASTQ on input (else we won't have quality information),
    # and we can't write PDB FASTA with secondary structure information
    # unless we have that on input.
    if saveAs == 'fastq' and not args.fastq:
        raise ValueError(
            'You have specified --saveAs fastq without using --fastq '
            'to indicate that the input is FASTQ. Please be explicit.')
    elif saveAs == 'fasta-ss' and not args.fasta_ss:
        raise ValueError(
            'You have specified --saveAs fasta-ss without using --fasta-ss '
            'to indicate that the input is PDB FASTA. Please be explicit.')

    write = sys.stdout.write
    kept = 0
    for read in reads:
        kept += 1
        write(read.toString(format_=saveAs))

    total = reads.unfilteredLength()

    if not args.quiet:
        print('Read %d sequence%s, kept %d (%.2f%%).' %
              (total, '' if total == 1 else 's', kept,
               0.0 if total == 0 else kept / total * 100.0), file=sys.stderr)



#!/usr/bin/env python

"""
Given a BLAST or DIAMOND JSON output files, the corresponding FASTA (or FASTQ)
sequence files, and filtering criteria, filter the reads according to given
criteria and write them to standard output.

Run with --help for help.
"""

from __future__ import print_function

import sys
import argparse
from itertools import chain

from dark.fasta import FastaReads
from dark.fastq import FastqReads


if __name__ == '__main__':

    # We do not use the addFASTACommandLineOptions and
    # parseFASTACommandLineOptions utility functions below because we allow
    # multiple FASTA or FASTQ files on the command line, which we specify
    # by --fasta and --fastq. And those names clash with the option names
    # used by those utility functions.

    parser = argparse.ArgumentParser(
        description='Filter reads alignments to select a subset of reads',
        epilog=('Given BLAST or DIAMOND JSON output files, the '
                'corresponding FASTA (or FASTQ) sequence files, and '
                'filtering  criteria, print a subset of matching reads.'))

    parser.add_argument(
        '--matcher', default='blast', choices=('blast', 'diamond'),
        help='The matching algorithm that was used to produce the JSON.')

    parser.add_argument(
        '--json', metavar='JSON-file', nargs='+', action='append',
        required=True, help='the JSON file(s) of BLAST or DIAMOND output.')

    # A mutually exclusive group for either FASTA or FASTQ files.
    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        '--fasta', metavar='FASTA-file', nargs='+', action='append',
        help=('the FASTA file(s) of sequences that were given to BLAST '
              'or DIAMOND.'))

    group.add_argument(
        '--fastq', metavar='FASTQ-file', nargs='+', action='append',
        help=('the FASTQ file(s) of sequences that were given to BLAST '
              'or DIAMOND.'))

    # Args specific to DIAMOND.

    # A group for either the DIAMOND FASTA file or a sqlite3 database
    # of the FASTA used to make the DIAMOND database.
    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        '--diamondDatabaseFastaFilename',
        help=('The filename of the FASTA file used to make the DIAMOND '
              'database. If --matcher diamond is used, either this argument '
              'or --diamondSqliteDatabaseFilename must be specified.'))

    group.add_argument(
        '--diamondSqliteDatabaseFilename',
        help=('The filename of the sqlite3 database file of FASTA metadata, '
              'made from the FASTA that was used to make the DIAMOND '
              'database. If --matcher diamond is used, either this argument '
              'or --diamondDatabaseFilename must be specified.'))

    parser.add_argument(
        '--diamondDatabaseFastaDirectory',
        help=('The directory where the FASTA file used to make the DIAMOND '
              'database can be found. This argument is only useful when '
              '--diamondSqliteDatabaseFilename is specified.'))

    # Args for filtering on ReadsAlignments.
    parser.add_argument(
        '--minStart', type=int, default=None,
        help='Reads that start before this subject offset should not be '
        'shown.')

    parser.add_argument(
        '--maxStop', type=int, default=None,
        help='Reads that end after this subject offset should not be shown.')

    parser.add_argument(
        '--oneAlignmentPerRead', default=False, action='store_true',
        help='If True, only keep the best alignment for each read.')

    parser.add_argument(
        '--maxAlignmentsPerRead', type=int, default=None,
        help=('Reads with more than this many alignments will be elided. Pass '
              'zero to only keep reads with no matches (alignments).'))

    parser.add_argument(
        '--scoreCutoff', type=float, default=None,
        help=('A float score. Matches with scores worse than this will be '
              'ignored.'))

    parser.add_argument(
        '--maxHspsPerHit', type=int, default=None,
        help='A numeric max number of HSPs to show for each hit on hitId.')

    parser.add_argument(
        '--whitelist', nargs='+', default=None, action='append',
        help='sequence titles that should be whitelisted')

    parser.add_argument(
        '--blacklist', nargs='+', default=None, action='append',
        help='sequence titles that should be blacklisted')

    parser.add_argument(
        '--titleRegex', default=None,
        help='a regex that sequence titles must match.')

    parser.add_argument(
        '--negativeTitleRegex', default=None,
        help='a regex that sequence titles must not match.')

    parser.add_argument(
        '--truncateTitlesAfter', default=None,
        help=('a string that titles will be truncated beyond. If the '
              'truncated version of a title has already been seen, '
              'that title will be skipped.'))

    parser.add_argument(
        '--minSequenceLen', type=int, default=None,
        help='sequences of lesser length will be elided.')

    parser.add_argument(
        '--maxSequenceLen', type=int, default=None,
        help='sequences of greater length will be elided.')

    parser.add_argument(
        '--taxonomy', default=None,
        help=('a string of the taxonomic group on which should be '
              'filtered. eg "Vira" will filter on viruses.'))

    args = parser.parse_args()

    # Flatten lists of lists that we get from using both nargs='+' and
    # action='append'. We use both because it allows people to use (e.g.)
    # --json on the command line either via "--json file1 --json file2" or
    # "--json file1 file2", or a combination of these. That way it's not
    # necessary to remember which way you're supposed to use it and you also
    # can't be hit by the subtle problem encountered in
    # https://github.com/acorg/dark-matter/issues/453
    jsonFiles = list(chain.from_iterable(args.json))
    whitelist = (
        set(chain.from_iterable(args.whitelist)) if args.whitelist else None)
    blacklist = (
        set(chain.from_iterable(args.blacklist)) if args.blacklist else None)

    # TODO: Add a --readClass command-line option in case we want to
    # process FASTA containing AA sequences.
    if args.fasta:
        reads = FastaReads(list(chain.from_iterable(args.fasta)))
    else:
        reads = FastqReads(list(chain.from_iterable(args.fastq)))

    if args.matcher == 'blast':
        from dark.blast.alignments import BlastReadsAlignments
        readsAlignments = BlastReadsAlignments(reads, jsonFiles)
    else:
        # Must be 'diamond' (due to parser.add_argument 'choices' argument).
        if (args.diamondDatabaseFastaFilename is None and
                args.diamondSqliteDatabaseFilename is None):
            print('Either --diamondDatabaseFastaFilename or '
                  '--diamondSqliteDatabaseFilename must be used with '
                  '--matcher diamond.', file=sys.stderr)
            sys.exit(1)
        elif not (args.diamondDatabaseFastaFilename is None or
                  args.diamondSqliteDatabaseFilename is None):
            print('--diamondDatabaseFastaFilename and '
                  '--diamondSqliteDatabaseFilename cannot both be used with '
                  '--matcher diamond.', file=sys.stderr)
            sys.exit(1)

        from dark.diamond.alignments import DiamondReadsAlignments
        readsAlignments = DiamondReadsAlignments(
            reads, jsonFiles,
            databaseFilename=args.diamondDatabaseFastaFilename,
            databaseDirectory=args.diamondDatabaseFastaDirectory,
            sqliteDatabaseFilename=args.diamondSqliteDatabaseFilename)

    readsAlignments.filter(
        maxAlignmentsPerRead=args.maxAlignmentsPerRead,
        minSequenceLen=args.minSequenceLen,
        maxSequenceLen=args.maxSequenceLen,
        minStart=args.minStart, maxStop=args.maxStop,
        oneAlignmentPerRead=args.oneAlignmentPerRead,
        maxHspsPerHit=args.maxHspsPerHit,
        scoreCutoff=args.scoreCutoff,
        whitelist=whitelist, blacklist=blacklist,
        titleRegex=args.titleRegex, negativeTitleRegex=args.negativeTitleRegex,
        truncateTitlesAfter=args.truncateTitlesAfter, taxonomy=args.taxonomy)

    format_ = 'fasta' if args.fasta else 'fastq'

    for readAlignments in readsAlignments:
        print(readAlignments.read.toStr(format_=format_))
